"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const multimethod_1 = require("@arrows/multimethod");
const json2csv_1 = require("json2csv");
const common_tags_1 = require("common-tags");
const flattenResults = (results) => {
    return results.map((result) => ({
        name: result.name,
        ops: result.ops,
        margin: result.margin,
        percentSlower: result.percentSlower,
        samples: result.samples,
        promise: result.promise,
        min: result.details.min,
        max: result.details.max,
        mean: result.details.mean,
        median: result.details.median,
        standardDeviation: result.details.standardDeviation,
        marginOfError: result.details.marginOfError,
        relativeMarginOfError: result.details.relativeMarginOfError,
        standardErrorOfMean: result.details.standardErrorOfMean,
        sampleVariance: result.details.sampleVariance,
    }));
};
const prepareJSON = (summary, options) => {
    const results = options.details
        ? summary.results
        : summary.results.map(({ name, ops, margin, percentSlower }) => {
            return { name, ops, margin, percentSlower };
        });
    const content = {
        name: summary.name,
        date: summary.date.toISOString(),
        version: options.version,
        results,
        fastest: summary.fastest,
        slowest: summary.slowest,
    };
    return JSON.stringify(content, null, 2);
};
const prepareCSV = (summary, options) => {
    const results = options.details
        ? flattenResults(summary.results)
        : summary.results.map(({ name, ops, margin, percentSlower }) => {
            return { name, ops, margin, percentSlower };
        });
    return (0, json2csv_1.parse)(results);
};
const prepareHTMLTable = (summary, options) => {
    const results = options.details
        ? flattenResults(summary.results)
        : summary.results.map(({ name, ops, margin, percentSlower }) => {
            return { name, ops, margin, percentSlower };
        });
    const headers = Object.keys(results[0])
        .map((key) => `<th>${key}</th>`)
        .join('');
    const rows = results
        .map((result) => {
        return (0, common_tags_1.stripIndent) `
      <tr>
        ${Object.values(result)
            .map((item) => (0, common_tags_1.html) `<td>${item}</td>`)
            .join('')}
      </tr>
      `;
    })
        .join('');
    const markup = (0, common_tags_1.html) `
    <!DOCTYPE html>
    <html lang="en">
      <head>
        <meta charset="UTF-8" />
        <meta name="viewport" content="width=device-width, initial-scale=1.0" />
        <meta http-equiv="X-UA-Compatible" />
        <title>${summary.name}</title>
      </head>
      <body>
        <table>
          <thead>
            <tr>
              ${headers}
            </tr>
          </thead>
          <tbody>
            ${rows}
          </tbody>
        </table>
      </body>
    </html>
  `;
    return (0, common_tags_1.stripIndent)(markup);
};
const prepareColors = (percents) => {
    return percents.map((percent) => {
        const hue = 120 * ((100 - percent) / 100);
        return `hsl(${hue}, 85%, 55%)`;
    });
};
const prepareHTMLChart = (summary) => {
    const labels = summary.results.map((result) => result.name);
    const values = summary.results.map((result) => result.ops);
    const percents = summary.results.map((result) => result.percentSlower);
    const suffix = summary.date.getTime();
    const markup = (0, common_tags_1.html) `
    <!DOCTYPE html>
    <html lang="en">
      <head>
        <meta charset="UTF-8" />
        <meta name="viewport" content="width=device-width, initial-scale=1.0" />
        <meta http-equiv="X-UA-Compatible" />
        <script src="https://cdn.jsdelivr.net/npm/chart.js@3.5.1/dist/chart.min.js"></script>
        <title>${summary.name}</title>
        <style>
          body {
            margin: 0;
            padding: 0;
            background: #ddd;
          }

          .container {
            box-sizing: border-box;
            height: 96vh;
            width: 96vw;
            margin: 2vh 2vw;
            resize: both;
            overflow: hidden;
            padding: 20px;
            background: white;
            box-shadow: 0 0 15px #aaa;
          }
        </style>
      </head>
      <body>
        <div class="container">
          <canvas id="chart${suffix}" width="16" height="9"></canvas>
        </div>
        <script>
          const format = (num) => {
            const [whole, fraction] = String(num).split('.')
            const chunked = []
            whole
              .split('')
              .reverse()
              .forEach((char, index) => {
                if (index % 3 === 0) {
                  chunked.unshift([char])
                } else {
                  chunked[0].unshift(char)
                }
              })

            const fractionStr = fraction !== undefined ? '.' + fraction : ''

            return (
              chunked.map((chunk) => chunk.join('')).join(' ') + fractionStr
            )
          }
          const ctx${suffix} = document
            .getElementById('chart${suffix}')
            .getContext('2d')
          const chart${suffix} = new Chart(ctx${suffix}, {
            type: 'bar',
            data: {
              labels: ${JSON.stringify(labels)},
              datasets: [
                {
                  data: ${JSON.stringify(values)},
                  backgroundColor: ${JSON.stringify(prepareColors(percents))},
                  borderColor: ${JSON.stringify(prepareColors(percents))},
                  borderWidth: 2,
                },
              ],
            },
            options: {
              maintainAspectRatio: false,
              plugins: {
                title: {
                  display: true,
                  text: '${summary.name}',
                  font: { size: 20 },
                  padding: 20,
                },
                legend: {
                  display: false,
                },
                tooltip: {
                  callbacks: {
                    label: (context) => {
                      return format(context.parsed.y) + ' ops/s'
                    },
                  },
                  displayColors: false,
                  backgroundColor: '#222222',
                  padding: 10,
                  cornerRadius: 5,
                  intersect: false,
                },
              },
              scales: {
                x: {
                  grid: {
                    color: '#888888',
                  },
                },
                y: {
                  title: {
                    display: true,
                    text: 'Operations per second',
                    padding: 10,
                  },
                  grid: {
                    color: '#888888',
                  },
                },
              },
            },
          })
        </script>
      </body>
    </html>
  `;
    return (0, common_tags_1.stripIndent)(markup);
};
const prepareFileContent = (0, multimethod_1.multi)((_, options) => options.format, (0, multimethod_1.method)('csv', prepareCSV), (0, multimethod_1.method)('table.html', prepareHTMLTable), (0, multimethod_1.method)('chart.html', prepareHTMLChart), (0, multimethod_1.method)(prepareJSON));
exports.default = prepareFileContent;
