"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.default = void 0;

var _path = _interopRequireDefault(require("path"));

var _lodash = require("lodash");

var _getExportedName = require("../utilities/getExportedName");

var _isIgnoredFilename = require("../utilities/isIgnoredFilename");

var _isIndexFile = require("../utilities/isIndexFile");

var _parseFilename = require("../utilities/parseFilename");

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

/**
 * @file Rule to ensure that filenames match the exports of the file
 * @author Stefan Lau
 * @see https://github.com/selaux/eslint-plugin-filenames/blob/32fc70dd7572211d1e5b97e06ec7a005c77fe8d4/lib/rules/match-exported.js
 */
const transformMap = {
  camel: _lodash.camelCase,
  kebab: _lodash.kebabCase,

  pascal(name) {
    return (0, _lodash.upperFirst)((0, _lodash.camelCase)(name));
  },

  snake: _lodash.snakeCase
};

const getStringToCheckAgainstExport = (parsed, replacePattern) => {
  const dirArray = parsed.dir.split(_path.default.sep);
  const lastDirectory = dirArray[dirArray.length - 1];

  if ((0, _isIndexFile.isIndexFile)(parsed)) {
    return lastDirectory;
  }

  return replacePattern ? parsed.name.replace(replacePattern, '') : parsed.name;
};

const getTransformsFromOptions = option => {
  const usedTransforms = option && option.push ? option : [option];
  return usedTransforms.map(name => {
    return transformMap[name];
  });
};

const transform = (exportedName, transforms) => {
  return transforms.map(format => {
    return format ? format(exportedName) : exportedName;
  });
};

const anyMatch = (expectedExport, transformedNames) => {
  return transformedNames.includes(expectedExport);
};

const getWhatToMatchMessage = transforms => {
  if (transforms.length === 1 && !transforms[0]) {
    return 'the exported name';
  }

  if (transforms.length > 1) {
    return 'any of the exported and transformed names';
  }

  return 'the exported and transformed name';
};

const create = context => {
  return {
    Program(node) {
      const transforms = getTransformsFromOptions(context.options[0]);
      const replacePattern = context.options[1] ? new RegExp(context.options[1], 'u') : null;
      const filename = context.getFilename();

      const absoluteFilename = _path.default.resolve(filename);

      const parsed = (0, _parseFilename.parseFilename)(absoluteFilename);
      const shouldIgnore = (0, _isIgnoredFilename.isIgnoredFilename)(filename);
      const exportedName = (0, _getExportedName.getExportedName)(node, context.options);
      const isExporting = Boolean(exportedName);
      const expectedExport = getStringToCheckAgainstExport(parsed, replacePattern);
      const transformedNames = transform(exportedName, transforms);
      const everythingIsIndex = exportedName === 'index' && parsed.name === 'index';
      const matchesExported = anyMatch(expectedExport, transformedNames) || everythingIsIndex;

      const reportIf = function (condition, messageForNormalFile, messageForIndexFile) {
        const message = !messageForIndexFile || !(0, _isIndexFile.isIndexFile)(parsed) ? messageForNormalFile : messageForIndexFile;

        if (condition) {
          context.report({
            data: {
              expectedExport,
              exportName: transformedNames.join('\', \''),
              extension: parsed.ext,
              name: parsed.base,
              whatToMatch: getWhatToMatchMessage(transforms)
            },
            message,
            node
          });
        }
      };

      if (shouldIgnore) {
        return;
      }

      reportIf(isExporting && !matchesExported, 'Filename \'{{expectedExport}}\' must match {{whatToMatch}} \'{{exportName}}\'.', 'The directory \'{{expectedExport}}\' must be named \'{{exportName}}\', after the exported value of its index file.');
    }

  };
};

var _default = {
  create,
  meta: {
    schema: [{
      anyOf: [{
        items: {
          type: 'string'
        },
        type: 'array'
      }, {
        type: 'string'
      }, {
        type: 'null'
      }],
      default: null
    }, {
      default: null,
      oneOf: [{
        type: 'string'
      }, {
        type: 'null'
      }]
    }, {
      default: false,
      type: 'boolean'
    }],
    type: 'suggestion'
  }
};
exports.default = _default;
module.exports = exports.default;